local _, addon = ...

-- functions
local pairs, setmetatable = pairs, setmetatable
local format = string.format
local FillLocalizedClassList = FillLocalizedClassList
local GetRealmName = GetRealmName
local UnitClass = UnitClass
local UnitFactionGroup = UnitFactionGroup
local UnitName = UnitName
-- constants
local RAID_CLASS_COLORS = RAID_CLASS_COLORS
-- frames
-- libs

function addon:GetRealmName()
    return string.gsub(GetRealmName(), "%s+", "")
end

function addon:GetPlayerInfo()
    return {
        realm = self:GetRealmName(),
        faction = UnitFactionGroup("player"),
        name = select(1, UnitName("player")),
        class = select(2, UnitClass("player")),
        level = UnitLevel("player"),
    }
end

function addon:GetCharacterKey(character)
    if character then
        return format("%s.%s.%s", character.realm, character.faction, character.name)
    end

    local realm = self:GetRealmName()
    local name = select(1, UnitName("player"))
    local faction = UnitFactionGroup("player")

    return format("%s.%s.%s", realm, faction, name)
end

function addon:GetCharacter(key)
    key = key or self:GetCharacterKey()
    if type(key) == "table" then return key end

    if not self.db.characters[key] then
        addon:Debug(format("Error: could not access character `%s`.", key or "(nil key)"))
        return
    end

    return self.db.characters[key]
end

function addon:GetCharacterKeystone(key)
    local character = self:GetCharacter(key)
    if not character then return end

    return character and character.keystone
end

-- TODO: Move elsewhere {
StaticPopupDialogs["BROKER_KEYSTONES_KEYS_RESET"] = {
    text = "Broker_Keystones has detected that your character doesn't have their |cffA335EE[%s %s]|r keystone anymore.|n|nDo you want to delete keystone data for your other characters?|n|n|cff999999(You can have this action performed automatically by going to the Broker_Keystone options.)|r",
    button1 = _G["YES"],
    button2 = _G["NO"],
    OnAccept = function()
        addon:Debug("Accepted reset prompt.")
        addon:DeleteAllCharacterKeystones()
        addon:DeleteAllCharacterRecords()
    end,
    OnCancel = function(_, reason)
        addon:Debug(format("Dismissed reset prompt: %s.", reason or "n/a"))
    end,
    timeout = 0,
    whileDead = true,
    hideOnEscape = true,
    preferredIndex = 3,
}

function addon:TriggerResetPrompt(...)
    addon:Debug("Triggering reset prompt...")

    local instanceName, level = ...
    if not instanceName then
        if addon.player.keystone then
        local _
        _, _, instanceName, level = self:GetKeystoneInformation(addon.player.keystone)
        else
            instanceName, level = "Debug", 2
        end
    end

    StaticPopup_Show("BROKER_KEYSTONES_KEYS_RESET", instanceName, level)
end
-- TODO: } Move elsewhere

function addon:SetCharacterKeystone(key, link)
    local character = self:GetCharacter(key)
    if not character then return end

    local existingKeystone = character.keystone
    if link ~= existingKeystone then
        character.keystone = link
    end

    if link and character.previousBestRun then
        character.previousBestRun = nil
    end

    if existingKeystone and not link then
        -- character had a keystone, now they don't
        if addon.options.autoResetKeystones then
            addon:Debug("Auto-accepted reset prompt.")
            addon:DeleteAllCharacterKeystones()
            addon:DeleteAllCharacterRecords()
        else
            local _, _, instanceName, level = self:GetKeystoneInformation(existingKeystone)
            -- StaticPopup_Show("BROKER_KEYSTONES_KEYS_RESET", instanceName, level)
            self:TriggerResetPrompt(instanceName, level)
        end
    end

    self:Fire(self.EVENT_KEYSTONE, key, link)
end

function addon:SetCharacterRecord(key, mapId, level)
    local character = self:GetCharacter(key)
    if not character then return end

    local existingBestRun = character.record
    if not mapId then
        if existingBestRun then
            character.previousBestRun = existingBestRun
        end
        character.record = nil
    else
        character.record = {mapId, level}
        character.previousBestRun = nil
    end

    self:Fire(self.EVENT_RECORD, key, mapId, level)
end

function addon:DeleteAllCharacterKeystones()
    for _, character in self:GetCharactersIterator() do
        if character.keystone ~= nil then
            self:SetCharacterKeystone(character, nil)
        end
    end
end

function addon:DeleteAllCharacterRecords()
    for _, character in self:GetCharactersIterator() do
        if character.record ~= nil then
            self:SetCharacterRecord(character, nil)
        end
    end
end

function addon:GetCharacters(filter)
    if not filter then
        return self.db.characters
    end

    local len = 0
    local t = {}
    for k, v in pairs(self.db.characters) do
        if filter(k, v) then
            len = len + 1
            t[k] = v
        end
    end

    return setmetatable(t, {
        __len = function() return len end,
    })
end

function addon:TableLength(t)
    local i = 0

    for x, y in pairs(t) do
        i = i + 1
    end

    return i
end

function addon:GetCharactersIterator(filter)
    return pairs(self:GetCharacters(filter))
end



function addon:GetCharacterName(key, colour, ignoreDifferentRealm)
    local character = addon:GetCharacter(key)
    if not character then return end

    if not ignoreDifferentRealm and character.realm ~= self.player.realm then
        if colour then
            return format("|c%s%s-%s|r", RAID_CLASS_COLORS[character.class].colorStr, character.name, character.realm)
        end

        return format("%s-%s", character.name, character.realm)
    end

    if colour then
        return format("|c%s%s|r", RAID_CLASS_COLORS[character.class].colorStr, character.name)
    end

    return character.name
end

function addon:GetFactionIcon(faction)
    local icon = "achievement_pvp_h_h"
    if faction == "Alliance" then
        icon = "achievement_pvp_a_a"
    end

    return format("|TInterface\\Icons\\%s:0:0:0:-1|t", icon)
end

function addon:GetFactionDisplayName(faction, colour)
    if colour then
        local hex = "FF3333"
        if faction == "Alliance" then
            hex = "2459FF"
        end

        return format("|cff%s%s|r", hex, faction)
    end

    return faction
end

function addon:GetCharacterDisplayName(key, colour, ignoreDifferentRealm)
    local character = self:GetCharacter(key)
    if not character then return end

    if character.faction ~= self.player.faction then
        return format("%s %s", self:GetFactionIcon(character.faction), self:GetCharacterName(key, colour, ignoreDifferentRealm))
    end

    return self:GetCharacterName(key, colour, ignoreDifferentRealm)
end

-- todo: cleanup
local _classNames = {}
FillLocalizedClassList(_classNames)
function addon:GetCharacterClassDisplayName(key, colour)
    local character = self:GetCharacter(key)
    if not character then return end

    return _classNames[character.class]
end

function addon:GetCharacterAnonymousDisplayName(key, colour)
    local character = self:GetCharacter(key)
    if not character then return end

    if character.faction ~= self.player.faction then
        return format("%s %s", self:GetFactionDisplayName(character.faction, colour), self:GetCharacterClassDisplayName(key, colour))
    end

    return self:GetCharacterClassDisplayName(key, colour)
end
